{-# LANGUAGE GeneralizedNewtypeDeriving
           , OverloadedStrings #-}

module Types where

import           Control.DeepSeq (NFData)
import           Data.ByteString (ByteString)
import qualified Data.ByteString.Char8 as BC
import           Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import           Data.Sequence (Seq)
import qualified Data.Sequence as Seq

import           IntBiMap (IntBiMap)
import qualified IntBiMap as IBM


newtype  SampleID = SampleID Int deriving (Eq, NFData, Ord)
type   LabelID = Int
type FeatureID = Int

type  Samples = [ SampleID]
type   Labels = [  LabelID]
type Features = [FeatureID]

newtype FeatureTable = FeatureTable (Seq Features)
type   LabelTable = Seq Labels
type   FeatureMap = IntBiMap ByteString
type     LabelMap = IntBiMap ByteString

type FeatureFreq = IntMap Int
type LabelFreq = IntMap Int

data LearnData = LearnData {
  labelTable   :: LabelTable
, featureTable :: FeatureTable
, labelMap     :: LabelMap
, featureMap   :: FeatureMap
, nPredictions :: Int
}

nAllLabels :: LearnData -> Int
nAllLabels = IBM.size . labelMap

nAllFeatures :: LearnData -> Int
nAllFeatures = IBM.size . featureMap

sampleFeatures :: FeatureTable -> SampleID -> Features
sampleFeatures (FeatureTable fT) (SampleID s) = Seq.index fT s

sampleLabels :: LabelTable -> SampleID -> Labels
sampleLabels lT (SampleID s) = Seq.index lT s

-- | Because of the isomorphism between labels and samples, we can convert
-- every sample to a label.
labelOfSample :: SampleID -> LabelID
labelOfSample (SampleID s) = s

sampleOfLabel :: LabelID -> SampleID
sampleOfLabel = SampleID

samplesRange :: SampleID -> SampleID -> Samples
samplesRange (SampleID from) (SampleID to) = map SampleID [from .. to - 1]

-- | Update feature frequency information with new samples.
featureFreq :: FeatureTable -> FeatureFreq -> Samples -> FeatureFreq
featureFreq fT fF =
  foldr (\ f -> IM.insertWith (+) f 1) fF . concatMap (sampleFeatures fT)


-- -----------------------------------------------------------------------------
-- Printing

showSample :: LabelMap -> SampleID -> ByteString
showSample m (SampleID s) = IBM.findElement m s

showLabel :: LabelMap -> LabelID -> ByteString
showLabel = IBM.findElement

showLabels :: LabelMap -> Labels -> ByteString
showLabels lM = BC.unwords . map (showLabel lM)

-- | Show a sample, followed by a list of labels.
showSampleLabels :: LabelMap -> SampleID -> Labels -> ByteString
showSampleLabels lM s ls = BC.concat [showSample lM s, ":", showLabels lM ls]
